#include "RNACons_Turner.h"
#include "Config.h"
#include "clusterRS.h"
#include "neighborjoin.h"

Config cf("Config/config"); 
TurnerE TE;

string infile, outfile="result.txt";//, logfile;

int knum = 5;
double delta = 15.0; //enumerate structures with scroe within 1.0 suboptimal of MFE+covariance
double barriercutoff = 12;
int percent = 20; //enumerate structures with free energies within 10% suboptimal of MFE;
bool is_countec = true; //true, count energy and covariance; false, only count energy; 
bool is_delta = false;
bool is_percent = true; 
bool reqLocOpt = true;
bool is_barriercutoff = false;


void print_usage() {
	cout<<"USAGE: RNAConSLOpt -i [infile] -o [outfile] -d delta -p percent -t barrier -c [configfile] "<<endl;
	cout<<"RNAConSLOpt can read in a multiple sequence alignment of a set of related RNAs and then produce all the consensus stable local optimal stack configurations within a certain energy range from the minimum free energy.  RNAConSLOpt can be seen as a combination of RNASLOpt and RNAalifold."<<endl;
	cout<<"  infile - a multiple sequence alignment of related RNAs"<<endl;
	cout<<"  delta  - the energy range is delta (kcal/mol)"<<endl;
	cout<<"  percent- the energy range is (1-percnet%)*MFE"<<endl;
	cout<<"  barrier- optional, the energy barrier cutoff"<<endl;
	cout<<"  configfile - optional, the configuration file"<<endl;
}

void init_env(int argc, char** argv){
	int opt = 0;
	while((opt=getopt(argc, argv, "i:o:c:d:b:p:t:"))!=EOF) {
		switch(opt) {
		case 'i': {infile  = string(optarg); break;}
		case 'o': {outfile = string(optarg); break;}
		case 'c': {string str(optarg); cf.setConfig(str); break;}
		case 'd': {string str(optarg); stringstream ss; ss<<str; ss>>delta; is_delta = true; break;}
		case 'p': {string str(optarg); stringstream ss; ss<<str; ss>>percent; is_percent = true; break;}
		case 't': {string str(optarg); stringstream ss(str); ss>>barriercutoff; is_barriercutoff=true; break;}
		default : {print_usage();exit(0);}
		}
	}
	if(infile.empty() ){
		print_usage(); exit(0);
	}
	if(outfile.empty()) {
		outfile = "result.txt";
	}
	if(is_delta && is_percent) {
		cerr<<"Both -d and -p are set, we choose to use -d only"<<endl; 
		is_percent = false;
	}

	cout<<"infile="<<infile<<", outfile="<<outfile<<endl;//", logfile="<<logfile<<endl;
}

//vvd is a lower triangular matrix, each entry (i,j) of which represents the conflicting degree between i and j
//cutoff is the new cut off, for each entry (i,j) with value less or equal to cutoff, we will merge column i and j to i, since the free energy of Si is always less or equal to Sj
//tree[i] = j means structrue Si is merged to Sj, if j==-1, Si is not merged. 
//cluster contains the indices of structures to consider. 
void applycutoff(vector <vector <double>> & vvd, double & cutoff, vector<TNode> & tree, vector<int> & cluster) {
	int n = sz(vvd);
	assert(n == sz(cluster));
	vector<int> newcluster;
	vector<vector<double>> newvvd;
	int m = 0;
	FOR(i,0,n) {
		bool merge = false;
		for(int j = sz(vvd[i])-1; j>=0; j--) {
			if(tree[j].pt == -1) {
				if(vvd[i][j] <= cutoff) {
					//merge j to i
					merge = true;
					tree[i].pt = j;
					break;
				}
			}
		}
		if(!merge) {
			m++; // the size of newcluster
			newcluster.PB(cluster[i]);
			assert(m==sz(newcluster));
			vector<double> vd;
			newvvd.PB(vd);
			assert(sz(newvvd)==m);
			FOR(j,0,m-1) {
				int k = newcluster[j];
				newvvd[m-1].PB(vvd[i][k]);
			}

		}
	}
	cluster = newcluster;
	vvd = newvvd;
}


int main(int argc, char** argv) {
	init_env(argc, argv);
	TE.init();

	ofstream fout(outfile.c_str()); 
	MSA org_msa(infile);

	fout<<"The original MSA:"<<endl;
	org_msa.show(fout);
	fout<<endl;
	
	vector<int> gap2ngap = org_msa.compute_gap2ngap(cf.P_MAXGAP);
	MSA msa = org_msa.apply_gap2ngap(gap2ngap);
	
	RNAConsT rc(msa);
	rc.init();
	rc.compute_covariance();
	rc.init_stl();
	
	int score = rc.stackfold();
	string mfe_str = rc.traceback();
	
	string mfe_str2 = reverse_gap2ngap(mfe_str, gap2ngap);
	fout<<"The minimum free energy structure:"<<endl<<mfe_str2<<endl<<endl;
	
	int e_ub = 0;
	if(is_delta && !is_percent)
		e_ub = score + rc.N * delta * UNIT;
	else if(is_percent && !is_delta) 
		e_ub = score * (100 - percent)/100;	

	fout<<"The predicted consensus local optimal stack configurations are as follows:"<<endl;
	vector<RNA> vrna = rc.generate_subopt(fout, e_ub, reqLocOpt);
	sort(vrna.begin(), vrna.end(), ascendingsortRNAbyE);
	
	if(sz(vrna)>0) {	
		FOR(t,0,sz(vrna)) {
			fout<<reverse_gap2ngap(vrna[t].str, gap2ngap)<<"\t"<<setw(6)<<(vrna[t].e/(double)rc.N/(double)100.0)<<" = "<<((vrna[t].e+vrna[t].c)/(double)rc.N/(double)100.0) << "+"<< (-vrna[t].c/(double)rc.N/(double)100.0) <<endl;
		}
	}
	fout<<"The number of consensus local optimal stack configurations is "<< sz(vrna)<<"."<<endl<<endl<<endl;	
	

	//Cluster all the suboptimal structures
	vector<vector<double>> vvd;
	//compute free energy of all possible stems in st.vstems.
	map<stem, int> stemEmap = computeStemE(rc, rc.stl.vstems); 
	//Heuristic
	vector<TNode> tree;
	vector<int> vrna_cluster;
	map<stemstem, vector<int>> CDSTmap;
	FOR(t,0,sz(rc.stl.vstems)) {
		FOR(k,0,sz(rc.stl.vstems)) {
			if(rc.stl.vstems[t].conflictwith(rc.stl.vstems[k]) && rc.stl.vstems[t]!=rc.stl.vstems[k]) {
				get_conflictingdegree_between_stems(CDSTmap, rc.stl.vstems[t], rc.stl.vstems[k]);
			}
		}
	}
	if(is_barriercutoff) {
		cf.BARRIERCUTOFF = barriercutoff;
	}
	
	FOR(t,0,sz(vrna)) {
		if(t%500==0 && t!=0) {
			cout<<"The algorithm has processed ["<<t<<"/"<<sz(vrna)<<"] consensus local optimal stack configurations."<<endl;
		}
		RStack rs_t(vrna[t].str);
		//compute the distance between vrna[t] and each structure in vrna_cluster, 
		//if vrna[t] is close to vrna[k=vrna_cluster[x]], then do nothing
		//otherwise, add t to vrna_cluster.
		//assume the minimum conflicting degree is greater than conflicting degree cut off.
		double min_e = cf.BARRIERCUTOFF+1;//min_e could be any value greater than cf.BARRIERCUTOFF
		vector<double> reverse_vd;
		int indx = -1;
		for(int x = sz(vrna_cluster)-1; x>=0; x--) {
			int k = vrna_cluster[x];
			RStack rs_k(vrna[k].str);
			double cdsum_tk = - compute_pairwise_conflictingdegree(rc, rs_t, rs_k, stemEmap, CDSTmap);
			double cdsum_kt = - compute_pairwise_conflictingdegree(rc, rs_k, rs_t, stemEmap, CDSTmap);
			double cdsum = min(cdsum_tk, cdsum_kt);
			reverse_vd.PB(cdsum);
			if(min_e > cdsum) {
				min_e = cdsum;
				indx = k;
			}
			if(min_e < cf.BARRIERCUTOFF) {
				//the conflicting degree between vrna[t] and vrna[indx is less than the cut off, then node[t] is merged to node[indx]
				tree.PB(indx);
				break;
			} else {
			}
		}
				
		if(min_e >= cf.BARRIERCUTOFF) {
				//the least conflicting degree between vrna[t] and any structure with indices vrna_cluster[.] is greater than the cut off 
				//t should be preserved.
				vrna_cluster.PB(t);
				//reverse_vd[x] contains the conflicting degree between vrna[t] and vrna[vrna_cluster[sz(reverse_vd)-1-x]]
				vector<double> vd;
				for(int x=sz(reverse_vd)-1; x>=0; x--) {
					vd.PB(reverse_vd[x]);
				}
				vvd.PB(vd);
				tree.PB(-1);//vrna[t] is not merged with any structure
			
				assert(sz(vrna_cluster) == sz(vd)+1 && sz(vvd)==sz(vd)+1);
				FOR(x,0,sz(vd)) {
					int k = vrna_cluster[x];
				}
			}
			
			if(sz(vrna_cluster) > cf.MAXMATRIXSZ) {
				cf.increaseBARRIERCUTOFF();
				cout<<"The number of consensus stable local optimal stack configurations generated is "<<sz(vrna_cluster)<<", while we can process no more than MAXMATRIXSZ("<<cf.MAXMATRIXSZ<<")."<<endl;
				cout<<"We will increase the energy barrier cutoff to "<<cf.BARRIERCUTOFF<<" kcal/mol in order to decrease the number of consensus stable local optimal structures."<<endl;
				applycutoff(vvd, cf.BARRIERCUTOFF, tree, vrna_cluster);
			}
		}
		
	
	vector<RNA> vrna_tmp;
	FOR(t,0,sz(vrna_cluster)) {
		vrna_tmp.PB(vrna[vrna_cluster[t]]);
	}
	
	fout<<"The predicted "<<sz(vrna_cluster)<<" consensus stable local optimal structures are as follows."<<endl;
	mytable mt (vvd);
	vector<int> vorder;
	vector<int> vi = mt.DJ(knum, vrna_tmp, fout, vorder);
		
	int t = sz(vorder)-1;
	while(t>=0) {
		t--;
	}

	t = sz(vorder)-1;
	while(t>=0) {
		fout<<vorder[t]<<"\t"<<reverse_gap2ngap(vrna_tmp[vorder[t]].str, gap2ngap)<<endl;
		t--;
	}

	fout<<endl<<endl<<"The top "<<sz(vi)<<" consensus stable local optimal stack configurations ranked according to their associated minimal energy barrier are as follows:"<<endl;
	FOR(t,0,sz(vi)) {
		fout<<reverse_gap2ngap(vrna_tmp[vi[t]].str, gap2ngap)<<endl;
	}
	fout<<endl<<endl;
	
	fout<<"Done"<<endl;
	fout.close();
	return 0;
}














